package office.api.main.sample3;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.io.File;
import java.io.StringWriter;
import java.net.URI;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.commons.cli.MissingArgumentException;
import org.apache.commons.httpclient.ConnectTimeoutException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import office.api.main.sample3.util.OptionsParser;

import office.api.base.BaseGetUsersById;
import office.api.base.BaseGetUsersByLoginName;
import office.api.common.CBServiceClient;
import office.api.main.sample3.util.Config;

public class UserInfo2Export {
	
	//ユーザーを識別する項目
	private static final String KEYITEM_ID     = "id";
	private static final String KEYITEM_NAME     = "name";
	
	//入力ファイルの文字コード
	private static final String UTF8 ="UTF-8";
	private static final String SHIFTJIS ="Shift-JIS";
	private static final String EUC ="EUC-JP";
	
	private static final String EXCEPTION_ERRMSG = "プログラム処理中にエラーが発生しました。処理を中止します";

	/**
	 * サイボウズOfficeからユーザー情報を取得するサンプルプログラムを実行します。
	 * 
	 * @param args
	 *            オプション
	 * @throws IOException
	 *             ファイル入出力時に問題が発生した場合
	 * @throws ParseException
	 *             引数にエラーがある場合
	 * @throws org.apache.commons.cli.ParseException 
	 * @throws MissingArgumentException 
	 */
	public static void main(String[] args) {
		//ログファイルの作成
		BufferedWriter logbw=null;
		try{
			FileOutputStream logfos = new FileOutputStream("./getusers.log",true);	//ログを追記モードで開く
			OutputStreamWriter logosw = new OutputStreamWriter(logfos,UTF8);
			logbw = new BufferedWriter(logosw);
		
			System.out.println(createMsg("処理を開始します"));
			writeLog(logbw, createMsg("処理を開始します"));
			
			//プロパティファイルの読み込み
			String propertiesFile = "./getusers.properties";
			String charset = "";
			String inputfile = "";
		
			// 引数チェック
			OptionsParser parser = null;
			try {
				parser = new OptionsParser(args);
			} catch (org.apache.commons.cli.ParseException e) {
				System.err.println(e.getMessage());
				writeLog(logbw,createMsg(e.getMessage()));
				logbw.newLine();
				logbw.close();
				return;
			}
		
			//プロパティファイル名の取得
			if (parser.hasPropertiesFileName())	
				propertiesFile = parser.getPropertiesFileName();
			
			//登録用CSVファイル名の取得
			if (parser.hasInputFileName())
				inputfile = parser.getInputFileName();
			else{
				System.err.println("Missing -i option");
				writeLog(logbw,createMsg("Missing -i option"));
				logbw.newLine();
				
				if(logbw != null)
					logbw.close();
				
				return;
			}
		
			//入力文字コードの設定
			if(parser.hasCharactorSet())
				charset = parser.getCharactorSet();
			else //-cを指定しない場合は初期値としてShift-JISを設定
				charset = SHIFTJIS;
		
			if(!charset.equals(UTF8) && !charset.equals(SHIFTJIS) && !charset.equals(EUC)){	//入力文字コードの設定ミス
				System.err.println("Missing -s option");
				writeLog(logbw,createMsg("Missing -s option"));
				logbw.newLine();
				if(logbw != null)
					logbw.close();
				
				return;
			}
		
			//プロパティファイルの読み込み
			System.out.println(createMsg("プロパティファイルを読み込みます"));
			writeLog(logbw,createMsg("プロパティファイルを読み込みます"));
	        Config config = null;
			try {
				config = new Config(propertiesFile);
			} catch (FileNotFoundException e) {
				System.err.println(createMsg("プロパティファイルが見つかりません"));
				writeLog(logbw,createMsg("プロパティファイルが見つかりません"));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			} catch (IOException e) {
				System.err.println(createMsg(e.getMessage()));
				writeLog(logbw,createMsg(e.getMessage()));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			}
			
			System.out.println(createMsg("プロパティファイルを読み込みました"));
			writeLog(logbw,createMsg("プロパティファイルを読み込みました"));
		
			String err = checkConfig(config);
		
			if(err.length() > 0){	//propetiesファイルの内容不備
				System.err.println(err);
				writeLog(logbw,createMsg(err));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			}
			
			//出力先ディレクトリの中身を空にする
			String exportdir = config.getExportDir();
			
			if(!exportdir.substring(exportdir.length() - 1).equals("/"))
				exportdir += "/";
			
			File dir = new File(exportdir);
			if(!dir.exists()){
				System.err.println(createMsg("出力先ディレクトリ(" + exportdir + ")が存在しません)"));
				writeLog(logbw,createMsg("出力先ディレクトリ(" + exportdir + ")が存在しません)"));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			}
			
			for(File file : dir.listFiles()){	//出力先ディレクトリ内のファイルをすべて削除
				file.delete();
			}
			System.out.println(createMsg("[" + exportdir + "] ディレクトリ内のファイルを削除しました"));
			writeLog(logbw,createMsg("[" + exportdir + "] ディレクトリ内のファイルを削除しました"));
			
			BufferedReader br = null;
			CBServiceClient client = null;
			
			int count = 0;
			int errcount = 0;
			int successcount = 0;
			
			//CSVファイルの読み込み
			try{
				FileInputStream fis = new FileInputStream(inputfile);
				InputStreamReader isr = new InputStreamReader(fis,charset);
				br = new BufferedReader(isr);
				System.out.println(createMsg("ファイルを開きました(" + inputfile + ")"));
				writeLog(logbw,createMsg("ファイルを開きました(" + inputfile + ")"));
			} catch (FileNotFoundException e) {
				System.err.println(createMsg(inputfile + "が見つかりません"));
				writeLog(logbw,createMsg(inputfile + "が見つかりません"));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			} catch (IOException e) {
				System.err.println(createMsg(e.getMessage()));
				writeLog(logbw,createMsg(e.getMessage()));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			}
			
			//SOAPクライアントのオブジェクト生成
			try {
				client = new CBServiceClient();
				client.load(config.getOfficeURL(),config.getUsername(),config.getPassword());
			} catch (AxisFault e) {
				System.err.println(e.getMessage());
				writeLog(logbw,createMsg(e.getMessage()));
				
				writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
				System.err.println(createMsg(EXCEPTION_ERRMSG));
				
				logbw.newLine();
				logbw.close();
				return;
			}
		
			String lines = "";
			
			List<String> errdata = new ArrayList<String>();
			
			int dataCnt = 0;
			List<String> inputDataList = new ArrayList<String>();
			
			try {
				while((lines = br.readLine()) != null){	//全データを取得し、データとデータ件数を取得する
					inputDataList.add(lines);
					dataCnt++;
				}
				//while((line = br.readLine()) != null){
				for(String data:inputDataList){
					count++;
					
					String[] columns = data.split(",");
										
					String loginID = "";				//ログインID
					
					String errMsgPrefix = count + "行目の処理中にエラーが発生しました(" 
											+ columns[0];
	
					if(config.getKeyItem().trim().equals(KEYITEM_NAME)){ //ユーザー識別項目にログイン名を指定している
						BaseGetUsersByLoginName baseAction = new BaseGetUsersByLoginName();
						baseAction.addLoginName(columns[0]);//ログイン名をセット
						
						try{
							OMElement result = client.sendReceive(baseAction);
							
							if(result != null){
								loginID = getLoginID(result);	//ログインIDを取得
				            }
							
						}catch(ConnectTimeoutException e){
							errdata.add(data);
							errcount++;
							System.err.println(createMsg("Officeへの接続に失敗しました。：接続タイムアウトエラー"));
							writeLog(logbw, createMsg("Officeへの接続に失敗しました。：接続タイムアウトエラー"));	
							return;
						}catch (AxisFault ae){
							errdata.add(data);
							errcount++;
							
							OMElement errCode = ae.getDetail();
							if(errCode != null){
								if(errCode.getText().equals("19106")){	//存在しないログイン名を指定している
									System.err.println(createMsg(errMsgPrefix + " 19106:ログイン名が不正です)"));
									writeLog(logbw, createMsg(errMsgPrefix + " 19106:ログイン名が不正です)"));
								}else if(errCode.getText().equals("19108")){	//接続ユーザーのID・パスワードエラー
									System.err.println(createMsg(errMsgPrefix + " 19108:officeへのログインに失敗しました)"));
									writeLog(logbw, createMsg(errMsgPrefix + " 19108:officeへのログインに失敗しました)"));
									return;
								}else{
									System.err.println(createMsg(errMsgPrefix + " " + errCode + ")"));
									writeLog(logbw, createMsg(errMsgPrefix + " " + errCode + ")"));
								}								
							}else{
								System.err.println(createMsg(errMsgPrefix + " " + ae.getMessage() + ")"));
								writeLog(logbw, createMsg(errMsgPrefix + " " + ae.getMessage() + ")"));
							}
							continue;			
						}catch(Throwable t){
							errcount++;
							errdata.add(data);
							System.err.println(t.getStackTrace());
							writeLog(logbw,createMsg(t.getMessage()));
							
							writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
							System.err.println(createMsg(EXCEPTION_ERRMSG));
							
							return;
						}
					}else{	//ユーザー識別項目にログインIDを指定している
						loginID = columns[0];
					}
					
					String exportFile = exportdir + columns[0] + ".xml";	//出力ファイル名（フルパス）
					
					FileOutputStream xmlfos = new FileOutputStream(exportFile,true);	//スケジュールデータ用オブジェクトを追記モードで生成
					OutputStreamWriter xmlsw = new OutputStreamWriter(xmlfos,UTF8);
					BufferedWriter xmlbw = new BufferedWriter(xmlsw);
					
					BaseGetUsersById action = new BaseGetUsersById();

					action.addUserID(loginID);	//ユーザーIDをセット	
					
					try{
						OMElement result = client.sendReceive(action);
						
						List<String> dataList = new ArrayList<String>();
						dataList = createUserData(result);
						xmlbw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
						xmlbw.newLine();
						
						for(String xmlData : dataList){
							xmlbw.write(xmlData);
							xmlbw.newLine();
						}
						xmlbw.close();		
						
						System.out.println(createMsg("ユーザー情報を出力しました。(" + exportFile + ")")); 
						writeLog(logbw,createMsg("ユーザー情報を出力しました。(" + exportFile + ")"));
						
					}catch(ConnectTimeoutException e){
						errcount++;
						errdata.add(data);
						System.err.println(createMsg("Officeへの接続に失敗しました。：接続タイムアウトエラー"));
						writeLog(logbw, createMsg("Officeへの接続に失敗しました。：接続タイムアウトエラー"));	
						
						writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
						System.err.println(createMsg(EXCEPTION_ERRMSG));
						
						return;
					}catch(AxisFault ae) {
						errcount++;
						errdata.add(data);
						
						OMElement errCode = ae.getDetail();
						if(errCode != null){
							if(errCode.getText().equals("19108")){	//接続ユーザーのID・パスワードエラー
								System.err.println(createMsg(errMsgPrefix + " 19108:ログイン名が不正です)"));
								writeLog(logbw, createMsg(errMsgPrefix + " 19108:ログイン名が不正です)"));
								return;
							}else{
								System.err.println(createMsg(errMsgPrefix + " Officeからのユーザー情報取得に失敗しました：" + ae.getMessage() + ")"));
								writeLog(logbw, createMsg(errMsgPrefix + " Officeからのユーザー情報取得に失敗しました：" + ae.getMessage() + ")"));
							}								
						}
						
						continue;
	
					}catch(Exception e){
						errcount++;
						errdata.add(data);
						System.err.println(createMsg(errMsgPrefix + " Officeからのユーザー情報取得に失敗しました：" + e.getMessage() + ")"));
						writeLog(logbw, createMsg(errMsgPrefix + " Officeからのユーザー情報取得に失敗しました：" + e.getMessage() + ")"));
						continue;					
					} catch (Throwable e) {
						errcount++;
						errdata.add(data);
						System.err.println(createMsg(e.getMessage()));
						writeLog(logbw,createMsg(e.getMessage()));
						
						writeLog(logbw, createMsg(EXCEPTION_ERRMSG));
						System.err.println(createMsg(EXCEPTION_ERRMSG));
						
						return;					
					}
					
					System.out.println(createMsg(count + "行目の処理が正常に終了しました(" 
											+ columns[0] + ")"));
					writeLog(logbw,createMsg(count + "行目の処理が正常に終了しました(" 
											+ columns[0] + ")"));
					
					successcount++;
				}
				
			}catch (IOException e) {
				System.err.println(createMsg("致命的なエラーが発生しました：" + e.getMessage()));
				writeLog(logbw, createMsg("致命的なエラーが発生しました：" + e.getMessage()));
			}finally{
				
				if(br != null){
					br.close();
					System.out.println(createMsg("ファイルを閉じました(" + inputfile + ")"));
					writeLog(logbw, createMsg("ファイルを閉じました(" + inputfile + ")"));
				}
				
				if(errdata.size() > 0){	//エラーデータが存在する
					BufferedWriter bw = null;
					FileOutputStream fos = new FileOutputStream("errdata.csv",false);	//エラーデータだけのCSVファイルを作成
					OutputStreamWriter osw = new OutputStreamWriter(fos,UTF8);
					bw = new BufferedWriter(osw);
					
					for(String errline : errdata){
						bw.write(errline);
						bw.newLine();
					}
					
					bw.close();
					
				}
				
				writeLog(logbw, createMsg("プログラムが終了しました(データ件数：" + dataCnt + "件 正常終了：" + successcount + "件 エラー：" + errcount + "件)"));
				System.out.println(createMsg("プログラムが終了しました(データ件数：" + dataCnt + "件 正常終了：" + successcount + "件 エラー：" + errcount + "件)"));

				logbw.newLine();
				
				if(logbw != null)
					logbw.close();
			}
		}catch (IOException e){
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	 /**
     * propertiesファイルの内容をチェックします。
     * 必須漏れや形式間違いがある場合はエラーメッセージを返します。
     * 
     * @param Config config propertiesファイル情報
     * @return String エラーメッセージ
     */
	private static String checkConfig(Config config){
		String errMsg = "";
		
		URI uri = config.getOfficeURL();
		String username = config.getUsername();
		String pwd = config.getPassword();
		String keyitem  = config.getKeyItem();
		String exportdir = config.getExportDir();
		
		boolean firstErrMsg = true;
		if(uri == null || uri.toString().trim().equals(""))
			errMsg = createErrMsg(Config.ConfigKeys.OFFICEURL.getKey(), firstErrMsg);
			
		if(username == null || username.trim().equals("")){
			if (errMsg.length() > 0)
				firstErrMsg = false;

			errMsg += createErrMsg(Config.ConfigKeys.USERNAME.getKey(), firstErrMsg);
		}
		
		if(pwd == null){
			if (errMsg.length() > 0)
				firstErrMsg = false;

			errMsg += createErrMsg(Config.ConfigKeys.PASSWORD.getKey(), firstErrMsg);
		}
		
		if(keyitem == null){
			if (errMsg.length() > 0)
				firstErrMsg = false;

			errMsg += createErrMsg(Config.ConfigKeys.KEYITEM.getKey(), firstErrMsg);
		}else{//設定値が「id」、「name」以外の場合
			if(!keyitem.equals(KEYITEM_ID) && !keyitem.equals(KEYITEM_NAME)){
				if (errMsg.length() > 0)
					firstErrMsg = false;

				errMsg += createErrMsg(Config.ConfigKeys.KEYITEM.getKey(), firstErrMsg);				
			}
		}
		
		if(exportdir == null){
			if (errMsg.length() > 0)
				firstErrMsg = false;

			errMsg += createErrMsg(Config.ConfigKeys.EXPORTDIR.getKey(), firstErrMsg);
		}
		

		return errMsg;
	}

	/**
     * propertiesファイルの内容チェック時でエラー時に出力されるメッセージを生成します。
     * 
     * @param String data エラーメッセージに表示させるチェック対象となるキー名
     * @param boolean firstMsg 最初のエラー可否のフラグ
     * @return String エラーメッセージ
     */
	private static String createErrMsg(String data,boolean firstMsg){
		String err = "";
		String crlf = System.getProperty("line.separator");
		if(data != null){
			if (!firstMsg){
				err += crlf;
			}
			err += "Value is not set correctly for " + data + " in properties file";
		}
		
		return err;
	}
	
	/**
     * ユーザーIDを取得します
     * 
     * @param OMElement node BaseGetUsersByLoginNameで取得されたノード情報
     * @return String ログインID
	 * @throws ParserConfigurationException
	 *             XMLの構成に問題が発生した場合
	 * @throws SAXException
	 *             XMLパース時に問題が発生した場合
	 * @throws IOException
	 *             ファイル入出力時に問題が発生した場合
	 * @throws XPathExpressionException
	 *             XPath式適用時に問題が発生した場合
     */
	private static String getLoginID(OMElement node) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException{
		
		Document document = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		document = builder.parse(new InputSource(new StringReader(node.toString())));

		String expression = "BaseGetUsersByLoginNameResponse/returns/user";

		XPath xPath = XPathFactory.newInstance().newXPath();
		XPathExpression xPathExpression = xPath.compile(expression);
		NodeList nodeList = (NodeList) xPathExpression.evaluate(document, XPathConstants.NODESET);
		
		String loginID = "";

		for (int i = 0; i < nodeList.getLength(); i++) {
			Node data = nodeList.item(i);
			NamedNodeMap map = data.getAttributes();
			
			for(int j = 0;j < map.getLength();j++){
				Node attrNode = map.item(j);
				if(attrNode.getNodeName().equals("key")){
					loginID = attrNode.getNodeValue();
					break;
				}
			}
		}
		return loginID;
	}
	
	/**
     * コンソール上に表示させるメッセージを生成します。
     * 
     * @param String msg 表示させる文字列
     * @param boolean success true:成功した処理 false:失敗した処理
     * @return String 生成したメッセージ
     * 
     */
	private static String createMsg(String msg){
		Date now = new Date();
		
		SimpleDateFormat formatter = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
		return "[" + formatter.format(now) + "] " + msg;
	}
	
	/**
     * ログファイルにデータを書き込みます
     * 
     * @param BufferedWriter bw ログファイルのBufferdWriterオブジェクト
     * @param String msg 表示させる文字列
  	 * @throws IOException
	 *             ファイル入出力時に問題が発生した場合
     * 
     */
	private static void writeLog(BufferedWriter bw,String msg) throws IOException{
		bw.write(msg);
		bw.newLine();
	}
	
	/**
     * SOAPで取得したデータからユーザーデータを作成します
     * 
     * @param OMElement node SOAPで取得したデータ
  	 * @throws ParserConfigurationException
  	 * @throws SAXException
  	 * @throws IOException
  	 * @throws XPathExpressionException
  	 * @throws TransformerException
  	 * @return List<String> 作成したユーザーデータ
     * 
     */
	private  static List<String> createUserData(OMElement node) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException, TransformerException{
		
		Document document = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		document = builder.parse(new InputSource(new StringReader(node.toString())));

		NodeList nodeList = null;
		String expression = "BaseGetUsersByIdResponse/returns/user";

		XPath xPath = XPathFactory.newInstance().newXPath();
		XPathExpression xPathExpression = xPath.compile(expression);
		nodeList = (NodeList) xPathExpression.evaluate(document, XPathConstants.NODESET);
		
		
		TransformerFactory tff=TransformerFactory.newInstance();
		Transformer tf = tff.newTransformer();
		tf.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,"yes");	//XML宣言を外す
		List<String> textList = new ArrayList<String>();
		
		StringWriter writer = new StringWriter();
		
		for (int i = 0; i < nodeList.getLength(); i++) {
			Node data = nodeList.item(i);
			tf.transform( new DOMSource(data),new StreamResult(writer));
			textList.add(writer.toString());
			writer = new StringWriter();
		}
		
		return textList;
	}
}
